@@ -147,6 +147,15 @@ crypto_create_session(struct fcrypt *fcr, struct session_op *sop)
 	case CRYPTO_AES_ECB:
 		alg_name = "ecb(aes)";
 		break;
+	case CRYPTO_AES_CFB:
+		alg_name = "cfb(aes)";
+		break;
+	case CRYPTO_AES_OFB:
+		alg_name = "ofb(aes)";
+		break;
+	case CRYPTO_AES_XTS:
+		alg_name = "xts(aes)";
+		break;
 	case CRYPTO_CAMELLIA_CBC:
 		alg_name = "cbc(camellia)";
 		break;
@@ -159,6 +168,21 @@ crypto_create_session(struct fcrypt *fcr, struct session_op *sop)
 		stream = 1;
 		aead = 1;
 		break;
+	case CRYPTO_AES_CCM:
+		alg_name = "ccm(aes)";
+		stream = 1;
+		aead = 1;
+		break;
+	case CRYPTO_TLS11_AES_CBC_HMAC_SHA1:
+		alg_name = "tls11(hmac(sha1),cbc(aes))";
+		stream = 0;
+		aead = 1;
+		break;
+	case CRYPTO_TLS12_AES_CBC_HMAC_SHA256:
+		alg_name = "tls12(hmac(sha256),cbc(aes))";
+		stream = 0;
+		aead = 1;
+		break;
 	case CRYPTO_NULL:
 		alg_name = "ecb(cipher_null)";
 		stream = 1;
@@ -424,6 +448,82 @@ crypto_get_session_by_sid(struct fcrypt *fcr, uint32_t sid)
 	return retval;
 }
 
+static void mutex_lock_double(struct mutex *a, struct mutex *b)
+{
+	if (b < a)
+		swap(a, b);
+
+	mutex_lock(a);
+	mutex_lock_nested(b, SINGLE_DEPTH_NESTING);
+}
+
+int
+crypto_get_sessions_by_sid(struct fcrypt *fcr,
+			   uint32_t sid_1, struct csession **ses_ptr_1,
+			   uint32_t sid_2, struct csession **ses_ptr_2)
+{
+	struct csession *ses_ptr;
+	int retval;
+
+	if (unlikely(fcr == NULL)) {
+		retval = -ENOENT;
+		goto out;
+	}
+
+	if (sid_1 == sid_2) {
+		retval = -EDEADLK;
+		goto out;
+	}
+
+	mutex_lock(&fcr->sem);
+
+	list_for_each_entry(ses_ptr, &fcr->list, entry) {
+		if (ses_ptr->sid == sid_1)
+			*ses_ptr_1 = ses_ptr;
+		else if (ses_ptr->sid == sid_2)
+			*ses_ptr_2 = ses_ptr;
+	}
+
+	if (*ses_ptr_1 && *ses_ptr_2) {
+		mutex_lock_double(&(*ses_ptr_1)->sem, &(*ses_ptr_2)->sem);
+		retval = 0;
+	} else {
+		retval = -ENOENT;
+	}
+
+	mutex_unlock(&fcr->sem);
+
+out:
+	if (retval) {
+		*ses_ptr_1 = NULL;
+		*ses_ptr_2 = NULL;
+	}
+	return retval;
+}
+
+#ifdef CIOCCPHASH
+/* Copy the hash state from one session to another */
+static int
+crypto_copy_hash_state(struct fcrypt *fcr, uint32_t dst_sid, uint32_t src_sid)
+{
+	struct csession *src_ses, *dst_ses;
+	int ret;
+
+	ret = crypto_get_sessions_by_sid(fcr, src_sid, &src_ses,
+					 dst_sid, &dst_ses);
+	if (unlikely(ret)) {
+		derr(1, "Failed to get sesssions with sid=0x%08X sid=%0x08X!",
+		     src_sid, dst_sid);
+		return ret;
+	}
+
+	ret = cryptodev_hash_copy(&dst_ses->hdata, &src_ses->hdata);
+	crypto_put_session(src_ses);
+	crypto_put_session(dst_ses);
+	return ret;
+}
+#endif /* CIOCCPHASH */
+
 static void cryptask_routine(struct work_struct *work)
 {
 	struct crypt_priv *pcr = container_of(work, struct crypt_priv, cryptask);
@@ -813,6 +913,9 @@ cryptodev_ioctl(struct file *filp, unsigned int cmd, unsigned long arg_)
 	struct crypt_priv *pcr = filp->private_data;
 	struct fcrypt *fcr;
 	struct session_info_op siop;
+#ifdef CIOCCPHASH
+	struct cphash_op cphop;
+#endif
 	uint32_t ses;
 	int ret, fd;
 
@@ -828,7 +931,13 @@ cryptodev_ioctl(struct file *filp, unsigned int cmd, unsigned long arg_)
 		fd = clonefd(filp);
 		ret = put_user(fd, p);
 		if (unlikely(ret)) {
+#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 17, 0))
 			sys_close(fd);
+#elif (LINUX_VERSION_CODE < KERNEL_VERSION(5, 11, 0))
+			ksys_close(fd);
+#else
+			close_fd(fd);
+#endif
 			return ret;
 		}
 		return ret;
@@ -859,6 +968,12 @@ cryptodev_ioctl(struct file *filp, unsigned int cmd, unsigned long arg_)
 		if (unlikely(ret))
 			return ret;
 		return copy_to_user(arg, &siop, sizeof(siop));
+#ifdef CIOCCPHASH
+	case CIOCCPHASH:
+		if (unlikely(copy_from_user(&cphop, arg, sizeof(cphop))))
+			return -EFAULT;
+		return crypto_copy_hash_state(fcr, cphop.dst_ses, cphop.src_ses);
+#endif /* CIOCPHASH */
 	case CIOCCRYPT:
 		if (unlikely(ret = kcop_from_user(&kcop, fcr, arg))) {
 			dwarning(1, "Error copying from user");
